local lg = love.graphics
local camera = {}

camera.x = 0
camera.y = 0
camera.velocity = Vec2(0)
camera.position = Vec2(player1.body:getPosition())
camera.lookOffset = Vec2(0)

camera.maxZoom = 2.5
camera.zoom = camera.maxZoom

-----------------------------

local function getScreenScale()
  return math.min(lg.getWidth()/1920, lg.getHeight()/1080)
end


function camera:moveTowards(targetVec, speed, dt)
  local minSpeed = 0.01
  speed = math.max(speed, minSpeed)

  local displacement = targetVec - self.position
  local direction = vector.normalize(displacement)

  local step = direction * speed * dt 



  if vector.getMagnitudeSqrd(step) >= vector.getMagnitudeSqrd(displacement) then

    self:setPosition(targetVec)
    
  else

    self:setPosition(self.position + step)
    
  end

end


function camera:getFinalZoom()
  return self.zoom*getScreenScale()
end


function camera:accellerateTowards(targetVel, accelleration, dt)
  
  local cameraPos = self.position
  
  local deltaV = targetVel - self.velocity
  
  
  local allignment = (1 + vector.dotProduct(vector.normalize(deltaV), vector.normalize(self.velocity)))*0.5
  
  local accelStep = deltaV * dt *accelleration * (2 - allignment)
  
  
  --Accellerate more when velocity is moving away
  
  if vector.getMagnitudeSqrd(accelStep) > vector.getMagnitudeSqrd(deltaV) then
    self.velocity = targetVel
  else
    self.velocity = self.velocity + accelStep
  end
  
end



function camera:moveZoom(targetZoom, speed, dt)

  if targetZoom - self.zoom ~= 0 then
    local direction = lume.sign(targetZoom - self.zoom)

    local minSpeed = 0.05
  
    
    speed = math.abs(speed)
    speed = math.max(speed, minSpeed)*direction
    

    local step = speed * dt
    

    local newZoom = self.zoom*1.07^step
    
    


    --If zoom goes over target, snap to it
    if (self.zoom <= targetZoom and targetZoom <= newZoom)
    or (self.zoom >= targetZoom and targetZoom >= newZoom) then
      self:setZoom(targetZoom)
    else
      self:setZoom(newZoom)
    end
  end
end


function camera:setPosition(x, y)
  if y then
    camera.position = Vec2(x, y)
  else
    camera.position = vector.clone(x)
  end
end

function camera:setZoom(zoom)
  self.zoom = zoom
end

function camera:setZoomSafe(zoom)
  self:setZoom(math.min(camera.maxZoom, zoom))
end

function camera:applyTransformations()
  lg.translate(window.width*0.5, window.height*0.5)
  lg.scale(self:getFinalZoom())
  lg.translate(-self.x, -self.y)
end


function camera:transformPoint(x, y)
  lg.push()
  self:applyTransformations()
  x, y = lg.transformPoint(x, y)
  lg.pop()
  return x, y
end


function camera:getPlayerDistance()
  return vector.getMagnitude(Vec2(player1.body:getPosition()) - self.position)
end

function camera:getPlayerDistanceSqrd()
  return vector.getMagnitudeSqrd(Vec2(player1.body:getPosition()) - self.position)
end


function camera:getLookVector()
  local lookVec

  if gamepadMode then
    local joystick = joysticks[1]
    lookVec = Vec2(joystick:getGamepadAxis("rightx"), joystick:getGamepadAxis("righty"))
    lookVec = vector.normalize(lookVec)
  else
    lookVec = Vec2(mouse.x, mouse.y)/ math.min(lg.getWidth(), lg.getHeight())*2
  end

  return lookVec
end


function camera:getLookPosition()
  
  local lookVec = self:getLookVector()
  local mult
  
  if player1.flightMode == "swooping" then
    mult = 0.4
  else
    mult = 0.3
  end
  
  local smallestDimension =  math.min(lg.getWidth(), lg.getHeight())
  local distance = (smallestDimension / self.zoom) * mult
  
  lookVec = lookVec * distance
  
  return lookVec
end


function camera:updateLookOffset(dt)
  
  local target = camera:getLookPosition()
  local displacement = target - self.lookOffset
  local speed  = 1.5/self.zoom
  local step = displacement*speed*dt
  
  if vector.getMagnitudeSqrd(step) >= vector.getMagnitudeSqrd(displacement) then
    self.lookOffset = vector.clone(target)
  else
    self.lookOffset = self.lookOffset + step
  end
  
end


function camera:getLookOffset()
  return self.lookOffset
end


function camera:getViewDimentions()
  return love.graphics.getWidth()/camera.zoom, love.graphics.getHeight()/camera.zoom
end

-----------------------------

local function getReactionDistance(speed)
  local reactionTime = 0.75
  return speed*reactionTime
end


local function getReactionZoom()
  local playerVel = Vec2(player1.body:getLinearVelocity())
  local speed     = vector.getMagnitude(playerVel)
  local distance  = getReactionDistance(speed)

  return math.min(lg.getHeight(), lg.getWidth())/distance
end


function camera:walkingCamera(dt)
  
  local speed = vector.getMagnitude(player1:getVelocity())
  
  local shiftOffset = Vec2(0, -40)
  local centrePos   = Vec2(player1.body:getPosition()) + shiftOffset

  local moveSpeed = vector.getMagnitude(centrePos - self.position)*2
  
  local targetZoom = camera.maxZoom
  local zoomVel = (targetZoom - self.zoom)*7
  
  local cameraA = vector.getMagnitude(player1:getVelocity() - self.velocity)*10
  self:moveZoom(targetZoom, zoomVel, dt)
  self:accellerateTowards(player1:getVelocity(), cameraA, dt)
  self:moveTowards(centrePos, moveSpeed, dt)
  
end


function camera:flightCamera(dt)

  local speed = vector.getMagnitude(player1:getVelocity())
  local centrePos = Vec2(player1.body:getPosition())

  local moveSpeed = self:getPlayerDistance()*2
  local targetZoom = math.min(getReactionZoom(), self.maxZoom)
  local zoomVel = (targetZoom - self.zoom)*7

  local cameraA = vector.getMagnitude(player1:getVelocity() - self.velocity)*100
  self:moveZoom(targetZoom, zoomVel, dt)
  self:accellerateTowards(player1:getVelocity(), cameraA, dt)
  self:moveTowards(centrePos, moveSpeed, dt)
  
end


function camera:zoomBehaviour()
  local foundPolygon = false
  if player1.flightMode == "swooping" and self.flyingTimer > self:getZoomChangeTime() then
    self.maxZoom = 1
  else
    for i, polygon in ipairs(polygonTable) do
      if polygon.attributes.zoomCamera > 0 then
        if polygon:checkInside(player1.body:getPosition()) then
          self.maxZoom = polygon.attributes.zoomCamera
          foundPolygon = true
          break
        end
      end
    end
    if not foundPolygon then
      self.maxZoom = 2
    end
  end
end


function camera:getZoomChangeTime()
  return 0.5
end


function camera:update(dt)
  
  self:zoomBehaviour(dt)

  local fastTucking = player1.tucked and vector.getMagnitudeSqrd(player1:getVelocity()) > 10^2
  local isFlying = player1.flightMode == "swooping"
  
  if isFlying then
    self.flyingTimer = self.flyingTimer + dt
  else
    self.flyingTimer = 0
  end
  
  if (isFlying and self.flyingTimer > 1)  or fastTucking then
    self:flightCamera(dt)
  else
    self:walkingCamera(dt)
  end
  
  self.position = self.position + self.velocity*dt
  
  self:updateLookOffset(dt)
  
  local finalPos = self.position + self:getLookOffset()
  
  self.x = finalPos.x
  self.y = finalPos.y
  
end




return camera 