WaterBlock = Prop:extend()
WaterBlock.type = "waterBlock"
local lg = love.graphics


local waterSources = require "tables.waterSources"


function WaterBlock:new(x, y, attributes)
  WaterBlock.super.new(self, x, y, attributes)
  
  createAttribute(self.attributes, "wetness", 10)
  
  createAttribute(self.attributes, "width", 1000)
  createAttribute(self.attributes, "height", 300)
  createAttribute(self.attributes, "rippleDensity", 0.1)
  

  self.attributes.rotation = nil
  self.attributes.animation = nil
  self.attributes.painted = nil
  self.attributes.textureName = nil
  self.attributes.scaleX = nil
  self.attributes.scaleY = nil
  
  self:addRipplePoints()
  
  table.insert(waterSources, self)
  
end


function WaterBlock:checkInside(x, y)
  
  local upperBound = self.y - self.height/2 + self:getLevelOffset()
  
  return self.x - self.width/2 < x and x <= self.x + self.width/2 and
         upperBound < y and y < self.y + self.height/2

end



function WaterBlock:delete()
  for i, v in ipairs(waterSources) do
    if v == self then
      table.remove(waterSources, i)
      break
    end
  end
  WaterBlock.super.delete(self)
end



function WaterBlock:checkSelect()
   
  return isPointInRectangle(
    mouse.worldX, 
    mouse.worldY, 
    self.x, 
    self.y, 
    self.width, 
    self.height)
end



local function getSign(i)
  
  if i > 0 then
    return 1
  else
    return -1
  end
  
end



function WaterBlock:getLevelOffset()
  
  if self.levelHeight then
    return math.max(self.levelHeight - (self.y - self:getHeight()*0.5), 0)
    
  else
    return 0
  end
end


local function newRipplePoint(parent)
  
  local self = {}
  

  self.y = 0
  self.vy = 0
  
  
  function self:playerInteract(x, y)
    
    if self.playerInteract then
      local pointPos = Vec2(x, y)
      local playerPos = player1:getPosition()
      local _, playerVelY = player1.body:getLinearVelocity()
      
      local radius = player1.shape:getRadius()
      
      if vector.getMagnitudeSqrd(pointPos - playerPos) < radius*radius then
        
        local mult = 400
        self.vy = math.tanh(0.3*playerVelY/mult)*mult
        
      end
    end
    
  end
  
  
  function self:getX()
    return self.x
  end
  
  function self:getY()
    return self.y
  end
  
  function self:getFinalY()
    return self:getY() + parent:getLevelOffset()
  end
  
  
  function self:update(x, dt, neighbours, ox, oy)
    
    local y = self:getY()
    local width = parent.attributes.width * smallTextureScale
    local spacing = width/(#parent.ripplePoints+1)
    local dragForce = -self.vy * 2
    local correctionForce = y^2 * 0.5 * getSign(-y) 
    local totalForce = dragForce + correctionForce
    
    
    for i, neighour in ipairs(neighbours) do
      local neighourForce = (neighour.y - y)*5*spacing
      totalForce = totalForce + neighourForce
    end
    
    self:playerInteract(x + ox, self:getFinalY() + oy)
    
    --Solve physics
    self.vy = self.vy + totalForce*dt
    self.y  = y + self.vy * dt
  end
  
  
  return self
  
end



function WaterBlock:getNumberOfVerts()
  return math.floor(self.attributes.width*self.attributes.rippleDensity)
end


function WaterBlock:addRipplePoints()
  self.ripplePoints = {}
  
  if self.attributes.rippleDensity > 0 then
    for i = 0, self:getNumberOfVerts()+1 do
      table.insert(self.ripplePoints, newRipplePoint(self))
    end
  end
end


function WaterBlock:getWidth()
  return self.attributes.width * smallTextureScale
end


function WaterBlock:getHeight()
  return self.attributes.height * smallTextureScale
end


function WaterBlock:getRippleVerts(offsetX, offsetY)
  --Generation of ripple verts goes from left to right
  --If attributes.rippleVerts is set to 0, there will be wtwo ripple verts for each corner.
  local vertTable = {}
  if self.attributes.rippleDensity > 0 then
    local width = self:getWidth()
    local height = self:getHeight()
    local spacing = width/(self:getNumberOfVerts()+1)
    
    
    
    local function insertVert(x, y)
      table.insert(vertTable, x + offsetX)
      table.insert(vertTable, math.min(y + offsetY, height+offsetY))
    end
    
    
    
    for i, ripplePoint in ipairs(self.ripplePoints) do
      insertVert(spacing*(i-1), ripplePoint:getFinalY())
    end
  end
  
  
  return vertTable
end


function WaterBlock:getPolygon()
  
  local width = self:getWidth()
  local height = self:getHeight()
  
  local vertTable = self:getRippleVerts(self.x - width/2, self.y - height/2)
  
  table.insert(vertTable, self.x + width/2)
  table.insert(vertTable, self.y + height/2)
  
  table.insert(vertTable, self.x - width/2)
  table.insert(vertTable, self.y + height/2)
  
  if not (self.attributes.rippleDensity > 0) then
    
    table.insert(vertTable, self.x - width/2)
    table.insert(vertTable, self.y - height/2)
    
    table.insert(vertTable, self.x + width/2)
    table.insert(vertTable, self.y - height/2)
  end
  
  
  return vertTable
end


function WaterBlock:updateBuilder(dt)
  
  self:setDimentions()
  
  if self.attributes.rippleVerts ~= #self.ripplePoints-2 then
    
    self:addRipplePoints()
  end
  
  self:grabAndDeleteBehaviour()
end


function WaterBlock:setDimentions()
  self.width = (self.attributes.width or 100) * smallTextureScale
  self.height = (self.attributes.height or 50) * smallTextureScale
end


function WaterBlock:playerInRange()
  
  local px, py = player1.body:getPosition()
  local radius = player1.shape:getRadius()
  
  
  local width = self:getWidth()
  local height = self:getHeight()
  
  local x = self.x
  local y = self.y - height*0.5
  
  
  self.playerInteract = x - width*0.5 <= px and px < x + width*0.5
                    and y - radius    <= py and py < y + radius
end


function WaterBlock:inSimulationDistance()
  
  local radius = self.width * 2
  
  return vector.getMagnitudeSqrd(Vec2(player1.body:getPosition()) - Vec2(self.x, self.y)) < radius*radius
  
end


function WaterBlock:setLevel(height)
  self.levelHeight = height
  cout(self:getLevelOffset())
end


function WaterBlock:update(dt)
  if self.attributes.rippleDensity > 0 then
    self:playerInRange()
    if self:inSimulationDistance() then
      for i, ripplePoint in ipairs(self.ripplePoints) do
        
        local width = self:getWidth()
        local height = self:getHeight()
        local spacing = width/(#self.ripplePoints+1)
        
        
        local offsetX = self.x - width/2
        local offsetY = self.y - height/2 
        
        local x = (i-1)*spacing 

        
     --[[   if self.levelHeight then 
          offsetY = math.max(self.levelHeight, offsetY)

          
          local y = math.max(y, self.levelHeight)
        end]]
        
        
        
        
        
        if not Entity.effectedByTimeBubble({getPosition = function() return x + offsetX, ripplePoint:getFinalY() end}) then
        
          local neighbours = {}
          
          --Add neighbours if they exist
          if self.ripplePoints[i-1] then
            table.insert(neighbours, self.ripplePoints[i-1])
          end
          if self.ripplePoints[i+1] then
            table.insert(neighbours, self.ripplePoints[i+1])
          end
          
          
          
          ripplePoint:update(x, dt, neighbours, offsetX, offsetY)
        end
      end
    end
  end
end



function WaterBlock:draw()

  if debugMode then
    local poly = self:getPolygon()
    lg.setColor(1,0,1)
    blue.graphics.polygon("line", poly)
    
    if self:inSimulationDistance() then
      for i, ripplePoint in ipairs(self.ripplePoints) do
        
        local width = self:getWidth()
        local height = self:getHeight()
        local spacing = width/(#self.ripplePoints+1)
        
        local offsetX = self.x - width/2
        local offsetY = self.y - height/2
        
        local x = (i-1)*spacing + offsetX
        local y = ripplePoint.y + offsetY
        

        
        blue.graphics.circle("line", x, y, 3)
        
      end
    end
    local width = self:getWidth()
    local height = self:getHeight()
    
    lg.setColor(1,1,0)
    local verts = self:getRippleVerts(self.x - width/2, self.y - height/2)
    for i = 1, #verts, 2 do
      
      local x = verts[i]
      local y = verts[i+1]

      
      blue.graphics.circle("line", x, y, 3)
      
    end
    
    
    
  end
end



function WaterBlock:pixelCanvasDraw()
  lg.push()
  camera:applyTransformations()
  
  local oldShader = love.graphics.getShader()
  love.graphics.setShader(shader.water)
  
  local width = self:getWidth()
  local height = self:getHeight()
  
  
  local poly = self:getPolygon()
  
  
  lg.setColor(1,1,1)
  
  if self.attributes.rippleDensity > 0 and not love.math.isConvex(poly) then
    local triangles = love.math.triangulate(poly)
    for i, triangle in ipairs(triangles) do
      love.graphics.polygon("fill", triangle)

    end
  else
    love.graphics.polygon("fill", poly)
  end
  
  love.graphics.setShader(oldShader)
  
  lg.pop()

end





