#define MAX_LIGHTS 32
#define MAX_AMBIENT_INTENSITY 4.0
#define GODRAY_MAX_DIFFERENCE 0.3

extern Image normalMap;
extern Image depthMap;
extern Image ambientMap;

extern vec3 outsideLight; //This is used to calculate godrays
extern vec3 lights[MAX_LIGHTS]; 
extern float lightIntensities[MAX_LIGHTS];
extern vec3 lightColours[MAX_LIGHTS];
extern bool flipX;
extern bool flipY;
extern float cameraZoom;
extern float rotation;

vec4 effect( vec4 color, Image texture, vec2 texture_coords, vec2 screen_coords )
{
  vec4 pixel = Texel(texture, texture_coords);

  
  //Since I don't want to use a different pixel format but want brightness higher than 1, I'll divide the intensity in set ambient and multiply it here.
  
  vec3 ambientLight = Texel(ambientMap, texture_coords).rgb * MAX_AMBIENT_INTENSITY;

  
  float screenZ = Texel(depthMap, texture_coords).r;
  
  vec4 normalPixel = Texel(normalMap, texture_coords);
  vec3 normal = normalPixel.rgb;
  normal = (normal-0.5)*2.0;


  //Get outside light similarity, this is to make godrays 'mist'
  vec3 difference = abs(ambientLight - outsideLight*MAX_AMBIENT_INTENSITY);
  float highestDifference = max(max(difference.r, difference.g), difference.b);

  if (highestDifference < GODRAY_MAX_DIFFERENCE)
  {
    float alpha = (1.0-(highestDifference / GODRAY_MAX_DIFFERENCE)) * 0.06;
    vec3 mist = vec3(1.0, 0.9686, 0.8196);
    pixel.rgb = pixel.rgb*(1.0-alpha) + mist*alpha;
    pixel.a = 1-((1.0-pixel.a)*(1.0-alpha));
  }

  if (pixel.r == 0.0 && pixel.g == 0.0 && pixel.b == 0.0)
  {
    discard;
    return pixel;
  }

  
  //Rotate normal
  
  
  float cosR = cos(rotation);
  float sinR = sin(rotation);
  
  normal = vec3(cosR*normal.x - sinR*normal.y, sinR*normal.x + cosR*normal.y, normal.z);
  
  
  
  if (flipX) normal.x = -normal.x;
  if (flipY) normal.y = -normal.y;
  
  //Now iterate through lights
  vec3 pointResult = vec3(0.0);
  vec3 averageColour = ambientLight;
  
  for (int i = 0; i < MAX_LIGHTS; i++)
  {
    
    float lightIntensity = lightIntensities[i];
    
    if (lightIntensity == 0.0) break; // If the light intensity is 0, assume that you reached the end of the array
    
    
    vec3 light          = lights[i];
    vec3 lightColour    = lightColours[i];
    light.z =  (screenZ - light.z) * cameraZoom;
    vec3 distance = light - vec3(screen_coords.x, screen_coords.y, 0.0);
    float lightFalloff = 1.0/pow(length(distance)/cameraZoom, 2.0);
    vec3 lightDirection = normalize(distance);
    float dotProduct = max((dot(lightDirection, normal))+0.1, 0.0);
    float finalBrightness = dotProduct * lightIntensity * lightFalloff * 2000.0;
    
    pointResult += pixel.rgb * lightColour  * finalBrightness;

    averageColour = averageColour + lightColour * finalBrightness;
    
  }
  
  //Make it so the average colour can't have any values greater than 1, but keep the ratio

  float highestValue = max(max(averageColour.r, averageColour.g), averageColour.b);

  averageColour = averageColour/highestValue;
  
  //Ambient light
  
  vec3 ambientResult = pixel.rgb * ambientLight;
  
  float alpha = pixel.a * color.a;
  
  vec3 lightResult = ambientResult + pointResult;
 
  float mult = 3.0;

  float highestChannel = max(max(lightResult.r, lightResult.g), lightResult.g);


  lightResult = min(pixel.rgb*mult*averageColour, lightResult);


  vec4 finalResult = vec4(lightResult/alpha, alpha);

  return finalResult;
}